@@ -19,16 +19,21 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
+import difflib
 import os
+import pathlib
+import tempfile
 import unittest
 from collections import defaultdict
 
 import yaml
+from lsst.resources import ResourcePath
 from pydantic import ValidationError
 
 from felis.datamodel import (
     CheckConstraint,
     Column,
+    ColumnGroup,
     Constraint,
     DataType,
     ForeignKeyConstraint,
@@ -39,8 +44,11 @@ from felis.datamodel import (
     UniqueConstraint,
 )
 
-TESTDIR = os.path.abspath(os.path.dirname(__file__))
-TEST_YAML = os.path.join(TESTDIR, "data", "test.yml")
+TEST_DIR = os.path.abspath(os.path.dirname(__file__))
+TEST_YAML = os.path.join(TEST_DIR, "data", "test.yml")
+TEST_SALES = os.path.join(TEST_DIR, "data", "sales.yaml")
+TEST_SERIALIZATION = os.path.join(TEST_DIR, "data", "test_serialization.yaml")
+TEST_ID_GENERATION = os.path.join(TEST_DIR, "data", "test_id_generation.yaml")
 
 
 class ColumnTestCase(unittest.TestCase):
@@ -256,6 +264,76 @@ class TableTestCase(unittest.TestCase):
             Table(name="testTable", id="#test_id", columns=[testCol, testCol])
 
 
+class ColumnGroupTestCase(unittest.TestCase):
+    """Test Pydantic validation of the ``ColumnGroup`` class."""
+
+    def test_validation(self) -> None:
+        """Test Pydantic validation of the ``ColumnGroup`` class."""
+        # Default initialization should throw an exception.
+        with self.assertRaises(ValidationError):
+            ColumnGroup()
+
+        # Setting only name should throw an exception.
+        with self.assertRaises(ValidationError):
+            ColumnGroup(name="testGroup")
+
+        # Setting name and id should throw an exception from missing columns.
+        with self.assertRaises(ValidationError):
+            ColumnGroup(name="testGroup", id="#test_id")
+
+        col = Column(name="testColumn", id="#test_col", datatype="string", length=256)
+
+        # Setting name, id, and columns should not throw an exception and
+        # should load data correctly.
+        group = ColumnGroup(name="testGroup", id="#test_group", columns=[col], ivoa_ucd="meta")
+        self.assertEqual(group.name, "testGroup", "name should be 'testGroup'")
+        self.assertEqual(group.id, "#test_group", "id should be '#test_group'")
+        self.assertEqual(group.columns, [col], "columns should be ['testColumn']")
+
+        # Dereferencing columns without setting a table should raise an
+        # exception.
+        with self.assertRaises(ValueError):
+            group._dereference_columns()
+
+        # Creating a group with duplicate column names should raise an
+        # exception.
+        with self.assertRaises(ValidationError):
+            ColumnGroup(name="testGroup", id="#test_group", columns=[col, col])
+
+        # Check that including a column object in a group works correctly.
+        group = ColumnGroup(name="testGroup", id="#test_group", columns=[col], ivoa_ucd="meta")
+        table = Table(
+            name="testTable",
+            id="#test_table",
+            columns=[col],
+            column_groups=[group],
+        )
+        self.assertEqual(table.column_groups, [group], "column_groups should be [group]")
+        self.assertEqual(col, table.column_groups[0].columns[0], "column_groups[0] should be testCol")
+
+        # Check that column derefencing works correctly when group is assigned
+        # to a table.
+        group = ColumnGroup(name="testGroup", id="#test_group", columns=["#test_col"], ivoa_ucd="meta")
+        table = Table(
+            name="testTable",
+            id="#test_table",
+            columns=[col],
+            column_groups=[group],
+        )
+        self.assertEqual(table.column_groups, [group], "column_groups should be [group]")
+        self.assertEqual(col, table.column_groups[0].columns[0], "column_groups[0] should be testCol")
+
+        # Creating a group with a bad column should raise an exception.
+        group = ColumnGroup(name="testGroup", id="#test_group", columns=["#bad_col"], ivoa_ucd="meta")
+        with self.assertRaises(ValueError):
+            table = Table(
+                name="testTable",
+                id="#test_table",
+                columns=[col],
+                column_groups=[group],
+            )
+
+
 class ConstraintTestCase(unittest.TestCase):
     """Test Pydantic validation of the different constraint classes."""
 
@@ -508,7 +586,7 @@ class SchemaTestCase(unittest.TestCase):
 
     def test_id_generation(self) -> None:
         """Test ID generation."""
-        test_path = os.path.join(TESTDIR, "data", "test_id_generation.yaml")
+        test_path = os.path.join(TEST_ID_GENERATION)
         with open(test_path) as test_yaml:
             yaml_data = yaml.safe_load(test_yaml)
             # Generate IDs for objects in the test schema.
@@ -519,6 +597,67 @@ class SchemaTestCase(unittest.TestCase):
             with self.assertRaises(ValidationError):
                 Schema.model_validate(yaml_data, context={"id_generation": False})
 
+    def test_get_table_by_column(self) -> None:
+        """Test the ``get_table_by_column`` method."""
+        # Test that the correct table is returned when searching by column.
+        test_col = Column(name="test_column", id="#test_tbl.test_col", datatype="string", length=256)
+        test_tbl = Table(name="test_table", id="#test_tbl", columns=[test_col])
+        sch = Schema(name="testSchema", id="#test_sch_id", tables=[test_tbl])
+        self.assertEqual(sch.get_table_by_column(test_col), test_tbl)
+
+        # Test that an error is raised when the column is not found.
+        bad_col = Column(name="bad_column", id="#test_tbl.bad_column", datatype="string", length=256)
+        with self.assertRaises(ValueError):
+            sch.get_table_by_column(bad_col)
+
+    def test_find_object_by_id(self) -> None:
+        test_col = Column(name="test_column", id="#test_tbl.test_col", datatype="string", length=256)
+        test_tbl = Table(name="test_table", id="#test_tbl", columns=[test_col])
+        sch = Schema(name="testSchema", id="#test_sch_id", tables=[test_tbl])
+        self.assertEqual(sch.find_object_by_id("#test_tbl.test_col", Column), test_col)
+        with self.assertRaises(KeyError):
+            sch.find_object_by_id("#bad_id", Column)
+        with self.assertRaises(TypeError):
+            sch.find_object_by_id("#test_tbl", Column)
+
+    def test_from_file(self) -> None:
+        """Test loading a schema from a file."""
+        # Test file object.
+        with open(TEST_SALES) as test_file:
+            schema = Schema.from_stream(test_file)
+            self.assertIsInstance(schema, Schema)
+
+        # Test path string.
+        schema = Schema.from_stream(open(TEST_SALES))
+        self.assertIsInstance(schema, Schema)
+
+        # Path object.
+        test_file_path = pathlib.Path(TEST_SALES)
+        schema = Schema.from_uri(test_file_path)
+        self.assertIsInstance(schema, Schema)
+
+    def test_from_resource(self) -> None:
+        """Test loading a schema from a resource."""
+        # Test loading a schema from a resource string.
+        schema = Schema.from_uri(
+            "resource://felis/schemas/tap_schema_std.yaml", context={"id_generation": True}
+        )
+        self.assertIsInstance(schema, Schema)
+
+        # Test loading a schema from a ResourcePath.
+        schema = Schema.from_uri(
+            ResourcePath("resource://felis/schemas/tap_schema_std.yaml"), context={"id_generation": True}
+        )
+        self.assertIsInstance(schema, Schema)
+
+        # Test loading from a nonexistant resource.
+        with self.assertRaises(ValueError):
+            Schema.from_uri("resource://fake/schemas/bad_schema.yaml")
+
+        # Without ID generation enabled, this schema should fail validation.
+        with self.assertRaises(ValidationError):
+            Schema.from_uri("resource://felis/schemas/tap_schema_std.yaml")
+
 
 class SchemaVersionTest(unittest.TestCase):
     """Test the schema version."""
@@ -744,5 +883,42 @@ class RedundantDatatypesTest(unittest.TestCase):
             Column(**{"name": "testColumn", "@id": "#test_col_id", "datatype": "double", "precision": 6})
 
 
+class SchemaSerializationTest(unittest.TestCase):
+    """Test serialization and deserialization of the schema data model."""
+
+    def test_serialization(self) -> None:
+        """Test serialization of the schema data model."""
+        # Read the original YAML content from the test_serialization.yaml file
+        with open(TEST_SERIALIZATION) as file:
+            original_yaml_content = file.read()
+
+        # Load the schema from the original YAML content
+        schema_out = Schema.from_uri(TEST_SERIALIZATION)
+        serialized_data = schema_out.model_dump(by_alias=True, exclude_none=True, exclude_defaults=True)
+
+        # Write the serialized data to a temporary YAML file
+        with tempfile.NamedTemporaryFile(delete=False, suffix=".yaml", mode="w+") as temp_file:
+            yaml.dump(serialized_data, temp_file, default_flow_style=False, sort_keys=False)
+            temp_file.seek(0)
+            # Read the deserialized YAML content from the temporary file
+            deserialized_yaml_content = temp_file.read()
+
+        # Show the differences between the original and deserialized YAML
+        diff = difflib.unified_diff(
+            original_yaml_content.splitlines(keepends=True),
+            deserialized_yaml_content.splitlines(keepends=True),
+            fromfile="original.yaml",
+            tofile="deserialized.yaml",
+        )
+        print("Differences:\n", "".join(diff))
+
+        # Assert that the original and deserialized YAML are the same
+        self.assertEqual(
+            yaml.safe_load(original_yaml_content),
+            yaml.safe_load(deserialized_yaml_content),
+            "The original and deserialized YAML contents should be the same",
+        )
+
+
 if __name__ == "__main__":
     unittest.main()
